package com.sun.xml.internal.fastinfoset.util;

import com.sun.xml.internal.fastinfoset.CommonResourceBundle;

public class StringIntMap extends KeyIntMap
{
    protected static final StringIntMap.Entry NULL_ENTRY;
    protected StringIntMap _readOnlyMap;
    protected StringIntMap.Entry _lastEntry;
    protected StringIntMap.Entry[] _table;
    protected int _index;
    protected int _totalCharacterCount;
    
    public StringIntMap(int initialCapacity, float loadFactor) {
        super(initialCapacity, loadFactor);
        this._lastEntry = StringIntMap.NULL_ENTRY;
        this._table = new StringIntMap.Entry[this._capacity];
    }
    
    public StringIntMap(int initialCapacity) {
        this(initialCapacity, 0.75f);
    }
    
    public StringIntMap() {
        this(16, 0.75f);
    }
    
    @Override
    public void clear() {
        for (int i = 0; i < this._table.length; ++i) {
            this._table[i] = null;
        }
        this._lastEntry = StringIntMap.NULL_ENTRY;
        this._size = 0;
        this._index = this._readOnlyMapSize;
        this._totalCharacterCount = 0;
    }
    
    @Override
    public void setReadOnlyMap(KeyIntMap readOnlyMap, boolean clear) {
        if (!(readOnlyMap instanceof StringIntMap)) {
            throw new IllegalArgumentException(CommonResourceBundle.getInstance().getString("message.illegalClass", new Object[] { readOnlyMap }));
        }
        this.setReadOnlyMap((StringIntMap)readOnlyMap, clear);
        return;
    }
    
    public final void setReadOnlyMap(StringIntMap readOnlyMap, boolean clear) {
        this._readOnlyMap = readOnlyMap;
        if (this._readOnlyMap != null) {
            this._readOnlyMapSize = this._readOnlyMap.size();
            this._index = this._size + this._readOnlyMapSize;
            if (clear) {
                this.clear();
            }
        }
        else {
            this._readOnlyMapSize = 0;
            this._index = this._size;
        }
    }
    
    public final int getNextIndex() {
        return this._index++;
    }
    
    public final int getIndex() {
        return this._index;
    }
    
    public final int obtainIndex(String key) {
        int hash = KeyIntMap.hashHash(key.hashCode());
        if (this._readOnlyMap != null) {
            int index = this._readOnlyMap.get(key, hash);
            if (index != -1) {
                return index;
            }
        }
        int tableIndex = KeyIntMap.indexFor(hash, this._table.length);
        StringIntMap.Entry e = this._table[tableIndex];
        while (e != null) {
            if (e._hash == hash && this.eq(key, e._key)) {
                return e._value;
            }
            e = e._next;
        }
        this.addEntry(key, hash, tableIndex);
        return -1;
    }
    
    public final void add(String key) {
        int hash = KeyIntMap.hashHash(key.hashCode());
        int tableIndex = KeyIntMap.indexFor(hash, this._table.length);
        this.addEntry(key, hash, tableIndex);
    }
    
    public final int get(String key) {
        if (key == this._lastEntry._key) {
            return this._lastEntry._value;
        }
        return this.get(key, KeyIntMap.hashHash(key.hashCode()));
    }
    
    public final int getTotalCharacterCount() {
        return this._totalCharacterCount;
    }
    
    private final int get(String key, int hash) {
        if (this._readOnlyMap != null) {
            int i = this._readOnlyMap.get(key, hash);
            if (i != -1) {
                return i;
            }
        }
        int tableIndex = KeyIntMap.indexFor(hash, this._table.length);
        StringIntMap.Entry e = this._table[tableIndex];
        while (e != null) {
            if (e._hash == hash && this.eq(key, e._key)) {
                this._lastEntry = e;
                return e._value;
            }
            e = e._next;
        }
        return -1;
    }
    
    private final void addEntry(String key, int hash, int bucketIndex) {
        StringIntMap.Entry e = this._table[bucketIndex];
        this._table[bucketIndex] = new StringIntMap.Entry(key, hash, this._index++, e);
        this._totalCharacterCount += key.length();
        if (this._size++ >= this._threshold) {
            this.resize(2 * this._table.length);
        }
    }
    
    protected final void resize(int newCapacity) {
        this._capacity = newCapacity;
        StringIntMap.Entry[] oldTable = this._table;
        int oldCapacity = oldTable.length;
        if (oldCapacity == 1048576) {
            this._threshold = Integer.MAX_VALUE;
            return;
        }
        StringIntMap.Entry[] newTable = new StringIntMap.Entry[this._capacity];
        this.transfer(newTable);
        this._table = newTable;
        this._threshold = (int)((float)this._capacity * this._loadFactor);
        return;
    }
    
    private final void transfer(StringIntMap.Entry[] newTable) {
        StringIntMap.Entry[] src = this._table;
        int newCapacity = newTable.length;
        for (int j = 0; j < src.length; ++j) {
            StringIntMap.Entry e = src[j];
            if (e != null) {
                src[j] = null;
                do {
                    StringIntMap.Entry next = e._next;
                    int i = KeyIntMap.indexFor(e._hash, newCapacity);
                    e._next = newTable[i];
                    newTable[i] = e;
                    e = next;
                } while (e != null);
            }
        }
    }
    
    private final boolean eq(String x, String y) {
        return x == y || x.equals(y);
    }
    
    static {
        NULL_ENTRY = new StringIntMap.Entry(null, 0, -1);
    }
    
    protected static class Entry extends KeyIntMap.BaseEntry
    {
        final String _key;
        StringIntMap.Entry _next;
        
        public Entry(String key, int hash, int value, StringIntMap.Entry next) {
            super(hash, value);
            this._key = key;
            this._next = next;
        }
    }
}
